@@ -31,6 +31,7 @@
 #include <string.h>
 #include <sys/socket.h>
 #include <arpa/inet.h>
+#include <netdb.h>
 
 #include "giognutls.h"
 #include "gresolv.h"
@@ -40,10 +41,20 @@
 
 #define SESSION_FLAG_USE_TLS	(1 << 0)
 
+enum chunk_state {
+	CHUNK_SIZE,
+	CHUNK_R_BODY,
+	CHUNK_N_BODY,
+	CHUNK_DATA,
+};
+
 struct _GWebResult {
 	guint16 status;
 	const guint8 *buffer;
 	gsize length;
+	gboolean use_chunk;
+	gchar *last_key;
+	GHashTable *headers;
 };
 
 struct web_session {
@@ -53,19 +64,30 @@ struct web_session {
 	char *host;
 	uint16_t port;
 	unsigned long flags;
+	struct addrinfo *addr;
 
 	char *content_type;
 
 	GIOChannel *transport_channel;
 	guint transport_watch;
+	guint send_watch;
 
 	guint resolv_action;
 	char *request;
 
 	guint8 *receive_buffer;
 	gsize receive_space;
+	GString *send_buffer;
 	GString *current_header;
 	gboolean header_done;
+	gboolean body_done;
+	gboolean more_data;
+	gboolean request_started;
+
+	enum chunk_state chunck_state;
+	gsize chunk_size;
+	gsize chunk_left;
+	gsize total_len;
 
 	GWebResult result;
 
@@ -83,8 +105,11 @@ struct _GWeb {
 	GList *session_list;
 
 	GResolv *resolv;
+	char *proxy;
 	char *accept_option;
 	char *user_agent;
+	char *user_agent_profile;
+	char *http_version;
 	gboolean close_connection;
 
 	GWebDebugFunc debug_func;
@@ -122,16 +147,32 @@ static void free_session(struct web_session *session)
 	if (session->transport_watch > 0)
 		g_source_remove(session->transport_watch);
 
+	if (session->send_watch > 0)
+		g_source_remove(session->send_watch);
+
 	if (session->transport_channel != NULL)
 		g_io_channel_unref(session->transport_channel);
 
-	g_string_free(session->current_header, TRUE);
+	g_free(session->result.last_key);
+
+	if (session->result.headers != NULL)
+		g_hash_table_destroy(session->result.headers);
+
+	if (session->send_buffer != NULL)
+		g_string_free(session->send_buffer, TRUE);
+
+	if (session->current_header != NULL)
+		g_string_free(session->current_header, TRUE);
+
 	g_free(session->receive_buffer);
 
 	g_free(session->content_type);
 
 	g_free(session->host);
 	g_free(session->address);
+	if (session->addr != NULL)
+		freeaddrinfo(session->addr);
+
 	g_free(session);
 }
 
@@ -200,8 +241,13 @@ void g_web_unref(GWeb *web)
 
 	g_resolv_unref(web->resolv);
 
+	g_free(web->proxy);
+
 	g_free(web->accept_option);
 	g_free(web->user_agent);
+	g_free(web->user_agent_profile);
+	g_free(web->http_version);
+
 	g_free(web);
 }
 
@@ -216,6 +262,24 @@ void g_web_set_debug(GWeb *web, GWebDebugFunc func, gpointer user_data)
 	g_resolv_set_debug(web->resolv, func, user_data);
 }
 
+gboolean g_web_set_proxy(GWeb *web, const char *proxy)
+{
+	if (web == NULL)
+		return FALSE;
+
+	g_free(web->proxy);
+
+	if (proxy == NULL) {
+		web->proxy = NULL;
+		debug(web, "clearing proxy");
+	} else {
+		web->proxy = g_strdup(proxy);
+		debug(web, "setting proxy %s", web->proxy);
+	}
+
+	return TRUE;
+}
+
 gboolean g_web_add_nameserver(GWeb *web, const char *address)
 {
 	if (web == NULL)
@@ -286,6 +350,37 @@ gboolean g_web_set_user_agent(GWeb *web, const char *format, ...)
 	return result;
 }
 
+gboolean g_web_set_ua_profile(GWeb *web, const char *profile)
+{
+	if (web == NULL)
+		return FALSE;
+
+	g_free(web->user_agent_profile);
+
+	web->user_agent_profile = g_strdup(profile);
+	debug(web, "setting user agent profile %s", web->user_agent);
+
+	return TRUE;
+}
+
+gboolean g_web_set_http_version(GWeb *web, const char *version)
+{
+	if (web == NULL)
+		return FALSE;
+
+	g_free(web->http_version);
+
+	if (version == NULL) {
+		web->http_version = NULL;
+		debug(web, "clearing HTTP version");
+	} else {
+		web->http_version = g_strdup(version);
+		debug(web, "setting HTTP version %s", web->http_version);
+	}
+
+	return TRUE;
+}
+
 void g_web_set_close_connection(GWeb *web, gboolean enabled)
 {
 	if (web == NULL)
@@ -304,13 +399,371 @@ gboolean g_web_get_close_connection(GWeb *web)
 
 static inline void call_result_func(struct web_session *session, guint16 status)
 {
+	gboolean result;
+
 	if (session->result_func == NULL)
 		return;
 
 	if (status != 0)
 		session->result.status = status;
 
-	session->result_func(&session->result, session->user_data);
+	result = session->result_func(&session->result, session->user_data);
+
+	debug(session->web, "[result function] %s",
+					result == TRUE ? "continue" : "stop");
+}
+
+static gboolean process_send_buffer(struct web_session *session)
+{
+	GString *buf = session->send_buffer;
+	gsize count, bytes_written;
+	GIOStatus status;
+
+	count = buf->len;
+
+	if (count == 0) {
+		if (session->request_started == TRUE &&
+					session->more_data == FALSE)
+			session->body_done = TRUE;
+
+		return FALSE;
+	}
+
+	debug(session->web, "bytes to write %zu", count);
+
+	status = g_io_channel_write_chars(session->transport_channel,
+					buf->str, count, &bytes_written, NULL);
+
+	debug(session->web, "status %u bytes written %zu",
+					status, bytes_written);
+
+	if (status != G_IO_STATUS_NORMAL && status != G_IO_STATUS_AGAIN)
+		return FALSE;
+
+	g_string_erase(buf, 0, bytes_written);
+
+	return TRUE;
+}
+
+static void process_next_chunk(struct web_session *session)
+{
+	GString *buf = session->send_buffer;
+	const guint8 *body;
+	gsize length;
+
+	if (session->input_func == NULL) {
+		session->more_data = FALSE;
+		return;
+	}
+
+	session->more_data = session->input_func(&body, &length,
+						session->user_data);
+
+	if (length > 0) {
+		g_string_append_printf(buf, "%zx\r\n", length);
+		g_string_append_len(buf, (char *) body, length);
+		g_string_append(buf, "\r\n");
+	}
+
+	if (session->more_data == FALSE)
+		g_string_append(buf, "0\r\n\r\n");
+}
+
+static void start_request(struct web_session *session)
+{
+	GString *buf = session->send_buffer;
+	const char *version;
+	const guint8 *body;
+	gsize length;
+
+	debug(session->web, "request %s from %s",
+					session->request, session->host);
+
+	g_string_truncate(buf, 0);
+
+	if (session->web->http_version == NULL)
+		version = "1.1";
+	else
+		version = session->web->http_version;
+
+	if (session->content_type == NULL)
+		g_string_append_printf(buf, "GET %s HTTP/%s\r\n",
+						session->request, version);
+	else
+		g_string_append_printf(buf, "POST %s HTTP/%s\r\n",
+						session->request, version);
+
+	g_string_append_printf(buf, "Host: %s\r\n", session->host);
+
+	if (session->web->user_agent != NULL)
+		g_string_append_printf(buf, "User-Agent: %s\r\n",
+						session->web->user_agent);
+
+	if (session->web->user_agent_profile != NULL) {
+		g_string_append_printf(buf, "x-wap-profile: %s\r\n",
+				       session->web->user_agent_profile);
+	}
+
+	if (session->web->accept_option != NULL)
+		g_string_append_printf(buf, "Accept: %s\r\n",
+						session->web->accept_option);
+
+	if (session->content_type != NULL) {
+		g_string_append_printf(buf, "Content-Type: %s\r\n",
+							session->content_type);
+		if (session->input_func == NULL) {
+			session->more_data = FALSE;
+			length = 0;
+		} else
+			session->more_data = session->input_func(&body, &length,
+							session->user_data);
+		if (session->more_data == FALSE)
+			g_string_append_printf(buf, "Content-Length: %zu\r\n",
+									length);
+		else
+			g_string_append(buf, "Transfer-Encoding: chunked\r\n");
+	}
+
+	if (session->web->close_connection == TRUE)
+		g_string_append(buf, "Connection: close\r\n");
+
+	g_string_append(buf, "\r\n");
+
+	if (session->content_type != NULL && length > 0) {
+		if (session->more_data == TRUE) {
+			g_string_append_printf(buf, "%zx\r\n", length);
+			g_string_append_len(buf, (char *) body, length);
+			g_string_append(buf, "\r\n");
+		} else
+			g_string_append_len(buf, (char *) body, length);
+	}
+}
+
+static gboolean send_data(GIOChannel *channel, GIOCondition cond,
+						gpointer user_data)
+{
+	struct web_session *session = user_data;
+
+	if (cond & (G_IO_NVAL | G_IO_ERR | G_IO_HUP)) {
+		session->send_watch = 0;
+		return FALSE;
+	}
+
+	if (process_send_buffer(session) == TRUE)
+		return TRUE;
+
+	if (session->request_started == FALSE) {
+		session->request_started = TRUE;
+		start_request(session);
+	} else if (session->more_data == TRUE)
+		process_next_chunk(session);
+
+	process_send_buffer(session);
+
+	if (session->body_done == TRUE) {
+		session->send_watch = 0;
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
+static int decode_chunked(struct web_session *session,
+					const guint8 *buf, gsize len)
+{
+	const guint8 *ptr = buf;
+	gsize counter;
+
+	while (len > 0) {
+		guint8 *pos;
+		gsize count;
+		char *str;
+
+		switch (session->chunck_state) {
+		case CHUNK_SIZE:
+			pos = memchr(ptr, '\n', len);
+			if (pos == NULL) {
+				g_string_append_len(session->current_header,
+						(gchar *) ptr, len);
+				return 0;
+			}
+
+			count = pos - ptr;
+			if (count < 1 || ptr[count - 1] != '\r')
+				return -EILSEQ;
+
+			g_string_append_len(session->current_header,
+						(gchar *) ptr, count);
+
+			len -= count + 1;
+			ptr = pos + 1;
+
+			str = session->current_header->str;
+
+			counter = strtoul(str, NULL, 16);
+			if ((counter == 0 && errno == EINVAL) ||
+						counter == ULONG_MAX)
+				return -EILSEQ;
+
+			session->chunk_size = counter;
+			session->chunk_left = counter;
+
+			session->chunck_state = CHUNK_DATA;
+			break;
+		case CHUNK_R_BODY:
+			if (*ptr != '\r')
+				return -EILSEQ;
+			ptr++;
+			len--;
+			session->chunck_state = CHUNK_N_BODY;
+			break;
+		case CHUNK_N_BODY:
+			if (*ptr != '\n')
+				return -EILSEQ;
+			ptr++;
+			len--;
+			session->chunck_state = CHUNK_SIZE;
+			break;
+		case CHUNK_DATA:
+			if (session->chunk_size == 0) {
+				debug(session->web, "Download Done in chunk");
+				g_string_truncate(session->current_header, 0);
+				return 0;
+			}
+
+			if (session->chunk_left <= len) {
+				session->result.buffer = ptr;
+				session->result.length = session->chunk_left;
+				call_result_func(session, 0);
+
+				len -= session->chunk_left;
+				ptr += session->chunk_left;
+
+				session->total_len += session->chunk_left;
+				session->chunk_left = 0;
+
+				g_string_truncate(session->current_header, 0);
+				session->chunck_state = CHUNK_R_BODY;
+				break;
+			}
+			/* more data */
+			session->result.buffer = ptr;
+			session->result.length = len;
+			call_result_func(session, 0);
+
+			session->chunk_left -= len;
+			session->total_len += len;
+
+			len -= len;
+			ptr += len;
+			break;
+		}
+	}
+
+	return 0;
+}
+
+static int handle_body(struct web_session *session,
+				const guint8 *buf, gsize len)
+{
+	int err;
+
+	debug(session->web, "[body] length %zu", len);
+
+	if (session->result.use_chunk == FALSE) {
+		if (len > 0) {
+			session->result.buffer = buf;
+			session->result.length = len;
+			call_result_func(session, 0);
+		}
+		return 0;
+	}
+
+	err = decode_chunked(session, buf, len);
+	if (err < 0) {
+		debug(session->web, "Error in chunk decode %d", err);
+
+		session->result.buffer = NULL;
+		session->result.length = 0;
+		call_result_func(session, 400);
+	}
+
+	return err;
+}
+
+static void handle_multi_line(struct web_session *session)
+{
+	gsize count;
+	char *str;
+	gchar *value;
+
+	str = session->current_header->str;
+
+	if (str[0] != ' ' && str[0] != '\t')
+		return;
+
+	while (str[0] == ' ' || str[0] == '\t')
+		str++;
+
+	count = str - session->current_header->str;
+	if (count > 0) {
+		g_string_erase(session->current_header, 0, count);
+		g_string_insert_c(session->current_header, 0, ' ');
+	}
+
+	value = g_hash_table_lookup(session->result.headers,
+					session->result.last_key);
+	if (value != NULL) {
+		g_string_insert(session->current_header, 0, value);
+
+		str = session->current_header->str;
+
+		g_hash_table_replace(session->result.headers,
+					g_strdup(session->result.last_key),
+					g_strdup(str));
+	}
+}
+
+static void add_header_field(struct web_session *session)
+{
+	gsize count;
+	guint8 *pos;
+	char *str;
+	gchar *value;
+	gchar *key;
+
+	str = session->current_header->str;
+
+	pos = memchr(str, ':', session->current_header->len);
+	if (pos != NULL) {
+		*pos = '\0';
+		pos++;
+
+		key = g_strdup(str);
+
+		/* remove preceding white spaces */
+		while (*pos == ' ')
+			pos++;
+
+		count = (char *) pos - str;
+
+		g_string_erase(session->current_header, 0, count);
+
+		value = g_hash_table_lookup(session->result.headers, key);
+		if (value != NULL) {
+			g_string_insert_c(session->current_header, 0, ' ');
+			g_string_insert_c(session->current_header, 0, ';');
+
+			g_string_insert(session->current_header, 0, value);
+		}
+
+		str = session->current_header->str;
+		g_hash_table_replace(session->result.headers, key,
+							g_strdup(str));
+
+		g_free(session->result.last_key);
+		session->result.last_key = g_strdup(key);
+	}
 }
 
 static gboolean received_data(GIOChannel *channel, GIOCondition cond,
@@ -335,7 +788,7 @@ static gboolean received_data(GIOChannel *channel, GIOCondition cond,
 
 	debug(session->web, "bytes read %zu", bytes_read);
 
-	if (status != G_IO_STATUS_NORMAL) {
+	if (status != G_IO_STATUS_NORMAL && status != G_IO_STATUS_AGAIN) {
 		session->transport_watch = 0;
 		session->result.buffer = NULL;
 		session->result.length = 0;
@@ -346,9 +799,11 @@ static gboolean received_data(GIOChannel *channel, GIOCondition cond,
 	session->receive_buffer[bytes_read] = '\0';
 
 	if (session->header_done == TRUE) {
-		session->result.buffer = session->receive_buffer;
-		session->result.length = bytes_read;
-		call_result_func(session, 0);
+		if (handle_body(session, session->receive_buffer,
+							bytes_read) < 0) {
+			session->transport_watch = 0;
+			return FALSE;
+		}
 		return TRUE;
 	}
 
@@ -375,13 +830,33 @@ static gboolean received_data(GIOChannel *channel, GIOCondition cond,
 						(gchar *) ptr, count);
 
 		bytes_read -= count + 1;
-		ptr = pos + 1;
+		if (bytes_read > 0)
+			ptr = pos + 1;
+		else
+			ptr = NULL;
 
 		if (session->current_header->len == 0) {
+			char *val;
+
 			session->header_done = TRUE;
-			session->result.buffer = pos + 1;
-			session->result.length = bytes_read;
-			call_result_func(session, 0);
+
+			val = g_hash_table_lookup(session->result.headers,
+							"Transfer-Encoding");
+			if (val != NULL) {
+				val = g_strrstr(val, "chunked");
+				if (val != NULL) {
+					session->result.use_chunk = TRUE;
+
+					session->chunck_state = CHUNK_SIZE;
+					session->chunk_left = 0;
+					session->total_len = 0;
+				}
+			}
+
+			if (handle_body(session, ptr, bytes_read) < 0) {
+				session->transport_watch = 0;
+				return FALSE;
+			}
 			break;
 		}
 
@@ -396,6 +871,12 @@ static gboolean received_data(GIOChannel *channel, GIOCondition cond,
 
 		debug(session->web, "[header] %s", str);
 
+		/* handle multi-line header */
+		if (str[0] == ' ' || str[0] == '\t')
+			handle_multi_line(session);
+		else
+			add_header_field(session);
+
 		g_string_truncate(session->current_header, 0);
 	}
 
@@ -404,42 +885,51 @@ static gboolean received_data(GIOChannel *channel, GIOCondition cond,
 
 static int connect_session_transport(struct web_session *session)
 {
-	struct sockaddr_in sin;
+	GIOFlags flags;
 	int sk;
 
-	sk = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
+	sk = socket(session->addr->ai_family, SOCK_STREAM, IPPROTO_TCP);
 	if (sk < 0)
 		return -EIO;
 
-	memset(&sin, 0, sizeof(sin));
-	sin.sin_family = AF_INET;
-	sin.sin_port = htons(session->port);
-	sin.sin_addr.s_addr = inet_addr(session->address);
-
-	if (connect(sk, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
-		close(sk);
-		return -EIO;
-	}
-
-	if (session->flags & SESSION_FLAG_USE_TLS)
+	if (session->flags & SESSION_FLAG_USE_TLS) {
+		debug(session->web, "using TLS encryption");
 		session->transport_channel = g_io_channel_gnutls_new(sk);
-	else
+	} else {
+		debug(session->web, "no encryption");
 		session->transport_channel = g_io_channel_unix_new(sk);
+	}
 
 	if (session->transport_channel == NULL) {
 		close(sk);
 		return -ENOMEM;
 	}
 
+	flags = g_io_channel_get_flags(session->transport_channel);
+	g_io_channel_set_flags(session->transport_channel,
+					flags | G_IO_FLAG_NONBLOCK, NULL);
+
 	g_io_channel_set_encoding(session->transport_channel, NULL, NULL);
 	g_io_channel_set_buffered(session->transport_channel, FALSE);
 
 	g_io_channel_set_close_on_unref(session->transport_channel, TRUE);
 
+	if (connect(sk, session->addr->ai_addr,
+			session->addr->ai_addrlen) < 0) {
+		if (errno != EINPROGRESS) {
+			close(sk);
+			return -EIO;
+		}
+	}
+
 	session->transport_watch = g_io_add_watch(session->transport_channel,
 				G_IO_IN | G_IO_HUP | G_IO_NVAL | G_IO_ERR,
 						received_data, session);
 
+	session->send_watch = g_io_add_watch(session->transport_channel,
+				G_IO_OUT | G_IO_HUP | G_IO_NVAL | G_IO_ERR,
+						send_data, session);
+
 	return 0;
 }
 
@@ -457,70 +947,8 @@ static int create_transport(struct web_session *session)
 	return 0;
 }
 
-static void start_request(struct web_session *session)
-{
-	GString *buf;
-	gchar *str;
-	const guint8 *body;
-	gsize length;
-	gsize count, bytes_written;
-	GIOStatus status;
-
-	debug(session->web, "request %s from %s",
-					session->request, session->host);
-
-	buf = g_string_new(NULL);
-
-	if (session->content_type == NULL)
-		g_string_append_printf(buf, "GET %s HTTP/1.1\r\n",
-							session->request);
-	else
-		g_string_append_printf(buf, "POST %s HTTP/1.1\r\n",
-							session->request);
-	g_string_append_printf(buf, "Host: %s\r\n", session->host);
-	if (session->web->user_agent != NULL)
-		g_string_append_printf(buf, "User-Agent: %s\r\n",
-						session->web->user_agent);
-	if (session->web->accept_option != NULL)
-		g_string_append_printf(buf, "Accept: %s\r\n",
-						session->web->accept_option);
-	if (session->content_type != NULL) {
-		g_string_append_printf(buf, "Content-Type: %s\r\n",
-						session->content_type);
-		if (session->input_func != NULL)
-			session->input_func(&body, &length, session->user_data);
-		else
-			length = 0;
-
-		g_string_append_printf(buf, "Content-Length: %zu\r\n", length);
-	}
-	if (session->web->close_connection == TRUE)
-		g_string_append(buf, "Connection: close\r\n");
-	g_string_append(buf, "\r\n");
-
-	count = buf->len;
-
-	if (session->content_type != NULL && length > 0) {
-		g_string_append_len(buf, (char *) body, length);
-		count += length;
-	}
-
-	str = g_string_free(buf, FALSE);
-
-	debug(session->web, "bytes to write %zu", count);
-
-	status = g_io_channel_write_chars(session->transport_channel,
-					str, count, &bytes_written, NULL);
-
-	debug(session->web, "status %u bytes written %zu",
-						status, bytes_written);
-
-	//printf("%s", str);
-
-	g_free(str);
-}
-
-static int parse_url(struct web_session *session, const char *url)
+static int parse_url(struct web_session *session,
+				const char *url, const char *proxy)
 {
 	char *scheme, *host, *port, *path;
 
@@ -551,7 +979,52 @@ static int parse_url(struct web_session *session, const char *url)
 	if (path != NULL)
 		*(path++) = '\0';
 
-	session->request = g_strdup_printf("/%s", path ? path : "");
+	if (proxy == NULL)
+		session->request = g_strdup_printf("/%s", path ? path : "");
+	else
+		session->request = g_strdup(url);
+
+	port = strrchr(host, ':');
+	if (port != NULL) {
+		char *end;
+		int tmp = strtol(port + 1, &end, 10);
+
+		if (*end == '\0') {
+			*port = '\0';
+			session->port = tmp;
+		}
+
+		if (proxy == NULL)
+			session->host = g_strdup(host);
+		else
+			session->host = g_strdup_printf("%s:%u", host, tmp);
+	} else
+		session->host = g_strdup(host);
+
+	g_free(scheme);
+
+	if (proxy == NULL)
+		return 0;
+
+	scheme = g_strdup(proxy);
+	if (scheme == NULL)
+		return -EINVAL;
+
+	host = strstr(proxy, "://");
+	if (host != NULL) {
+		*host = '\0';
+		host += 3;
+
+		if (strcasecmp(scheme, "http") != 0) {
+			g_free(scheme);
+			return -EINVAL;
+		}
+	} else
+		host = scheme;
+
+	path = strchr(host, '/');
+	if (path != NULL)
+		*(path++) = '\0';
 
 	port = strrchr(host, ':');
 	if (port != NULL) {
@@ -564,7 +1037,7 @@ static int parse_url(struct web_session *session, const char *url)
 		}
 	}
 
-	session->host = g_strdup(host);
+	session->address = g_strdup(host);
 
 	g_free(scheme);
 
@@ -575,6 +1048,8 @@ static void resolv_result(GResolvResultStatus status,
 					char **results, gpointer user_data)
 {
 	struct web_session *session = user_data;
+	struct addrinfo hints;
+	int ret;
 
 	if (results == NULL || results[0] == NULL) {
 		call_result_func(session, 404);
@@ -583,7 +1058,16 @@ static void resolv_result(GResolvResultStatus status,
 
 	debug(session->web, "address %s", results[0]);
 
-	if (inet_aton(results[0], NULL) == 0) {
+	memset(&hints, 0, sizeof(struct addrinfo));
+	hints.ai_flags = AI_NUMERICHOST;
+
+	if (session->addr != NULL) {
+		freeaddrinfo(session->addr);
+		session->addr = NULL;
+	}
+
+	ret = getaddrinfo(results[0], NULL, &hints, &session->addr);
+	if (ret != 0 || session->addr == NULL) {
 		call_result_func(session, 400);
 		return;
 	}
@@ -594,8 +1078,6 @@ static void resolv_result(GResolvResultStatus status,
 		call_result_func(session, 409);
 		return;
 	}
-
-	start_request(session);
 }
 
 static guint do_request(GWeb *web, const char *url,
@@ -613,13 +1095,16 @@ static guint do_request(GWeb *web, const char *url,
 	if (session == NULL)
 		return 0;
 
-	if (parse_url(session, url) < 0) {
+	if (parse_url(session, url, web->proxy) < 0) {
 		free_session(session);
 		return 0;
 	}
 
-	debug(web, "host %s:%u", session->host, session->port);
+	debug(web, "address %s", session->address);
+	debug(web, "port %u", session->port);
+	debug(web, "host %s", session->host);
 	debug(web, "flags %lu", session->flags);
+	debug(web, "request %s", session->request);
 
 	if (type != NULL) {
 		session->content_type = g_strdup(type);
@@ -639,11 +1124,20 @@ static guint do_request(GWeb *web, const char *url,
 		return 0;
 	}
 
+	session->result.headers = g_hash_table_new_full(g_str_hash, g_str_equal,
+							g_free, g_free);
+	if (session->result.headers == NULL) {
+		free_session(session);
+		return 0;
+	}
+
 	session->receive_space = DEFAULT_BUFFER_SIZE;
+	session->send_buffer = g_string_sized_new(0);
 	session->current_header = g_string_sized_new(0);
 	session->header_done = FALSE;
+	session->body_done = FALSE;
 
-	if (inet_aton(session->host, NULL) == 0) {
+	if (session->address == NULL && inet_aton(session->host, NULL) == 0) {
 		session->resolv_action = g_resolv_lookup_hostname(web->resolv,
 					session->host, resolv_result, session);
 		if (session->resolv_action == 0) {
@@ -651,14 +1145,13 @@ static guint do_request(GWeb *web, const char *url,
 			return 0;
 		}
 	} else {
-		session->address = g_strdup(session->host);
+		if (session->address == NULL)
+			session->address = g_strdup(session->host);
 
 		if (create_transport(session) < 0) {
 			free_session(session);
 			return 0;
 		}
-
-		start_request(session);
 	}
 
 	web->session_list = g_list_append(web->session_list, session);
@@ -712,6 +1205,23 @@ gboolean g_web_result_get_chunk(GWebResult *result,
 	return TRUE;
 }
 
+gboolean g_web_result_get_header(GWebResult *result,
+				const char *header, const char **value)
+{
+	if (result == NULL)
+		return FALSE;
+
+	if (value == NULL)
+		return FALSE;
+
+	*value = g_hash_table_lookup(result->headers, header);
+
+	if (*value == NULL)
+		return FALSE;
+
+	return TRUE;
+}
+
 struct _GWebParser {
 	gint ref_count;
 	char *begin_token;
