@@ -40,7 +40,6 @@ struct connman_network {
 	connman_uint8_t strength;
 	connman_uint16_t frequency;
 	char *identifier;
-	char *address;
 	char *name;
 	char *node;
 	char *group;
@@ -67,6 +66,9 @@ struct connman_network {
 		char *private_key_path;
 		char *private_key_passphrase;
 		char *phase2_auth;
+		connman_bool_t wps;
+		connman_bool_t use_wps;
+		char *pin_wps;
 	} wifi;
 };
 
@@ -168,11 +170,11 @@ static void network_destruct(struct connman_element *element)
 	g_free(network->wifi.private_key_path);
 	g_free(network->wifi.private_key_passphrase);
 	g_free(network->wifi.phase2_auth);
+	g_free(network->wifi.pin_wps);
 
 	g_free(network->group);
 	g_free(network->node);
 	g_free(network->name);
-	g_free(network->address);
 	g_free(network->identifier);
 
 	network->device = NULL;
@@ -436,7 +438,7 @@ connman_bool_t __connman_network_get_weakness(struct connman_network *network)
 	return FALSE;
 }
 
-connman_bool_t __connman_network_get_connecting(struct connman_network *network)
+connman_bool_t connman_network_get_connecting(struct connman_network *network)
 {
 	return network->connecting;
 }
@@ -497,7 +499,8 @@ int connman_network_set_associating(struct connman_network *network,
 
 		service = __connman_service_lookup_from_network(network);
 		__connman_service_indicate_state(service,
-					CONNMAN_SERVICE_STATE_ASSOCIATION);
+					CONNMAN_SERVICE_STATE_ASSOCIATION,
+					CONNMAN_IPCONFIG_TYPE_IPV4);
 	}
 
 	return 0;
@@ -515,7 +518,8 @@ static void set_associate_error(struct connman_network *network)
 	service = __connman_service_lookup_from_network(network);
 
 	__connman_service_indicate_state(service,
-					CONNMAN_SERVICE_STATE_FAILURE);
+					CONNMAN_SERVICE_STATE_FAILURE,
+					CONNMAN_IPCONFIG_TYPE_IPV4);
 }
 
 static void set_configure_error(struct connman_network *network)
@@ -527,7 +531,18 @@ static void set_configure_error(struct connman_network *network)
 	service = __connman_service_lookup_from_network(network);
 
 	__connman_service_indicate_state(service,
-					CONNMAN_SERVICE_STATE_FAILURE);
+					CONNMAN_SERVICE_STATE_FAILURE,
+					CONNMAN_IPCONFIG_TYPE_IPV4);
+}
+
+static void set_invalid_key_error(struct connman_network *network)
+{
+	struct connman_service *service;
+
+	service = __connman_service_lookup_from_network(network);
+
+	__connman_service_indicate_error(service,
+					CONNMAN_SERVICE_ERROR_INVALID_KEY);
 }
 
 void connman_network_set_ipv4_method(struct connman_network *network,
@@ -584,9 +599,28 @@ void connman_network_set_error(struct connman_network *network,
 	case CONNMAN_NETWORK_ERROR_CONFIGURE_FAIL:
 		set_configure_error(network);
 		break;
+	case CONNMAN_NETWORK_ERROR_INVALID_KEY:
+		set_invalid_key_error(network);
+		break;
 	}
 }
 
+void connman_network_clear_error(struct connman_network *network)
+{
+	struct connman_service *service;
+
+	DBG("network %p", network);
+
+	if (network == NULL)
+		return;
+
+	if (network->connecting == TRUE || network->associating == TRUE)
+		return;
+
+	service = __connman_service_lookup_from_network(network);
+	__connman_service_clear_error(service);
+}
+
 static void set_configuration(struct connman_network *network)
 {
 	struct connman_service *service;
@@ -601,54 +635,112 @@ static void set_configuration(struct connman_network *network)
 
 	service = __connman_service_lookup_from_network(network);
 	__connman_service_indicate_state(service,
-					CONNMAN_SERVICE_STATE_CONFIGURATION);
+					CONNMAN_SERVICE_STATE_CONFIGURATION,
+					CONNMAN_IPCONFIG_TYPE_IPV4);
+}
+
+static void dhcp_success(struct connman_network *network)
+{
+	struct connman_service *service;
+	struct connman_ipconfig *ipconfig_ipv4;
+	int err;
+
+	service = __connman_service_lookup_from_network(network);
+	if (service == NULL)
+		goto err;
+
+	connman_network_set_associating(network, FALSE);
+
+	network->connecting = FALSE;
+
+	ipconfig_ipv4 = __connman_service_get_ip4config(service);
+	err = __connman_ipconfig_address_add(ipconfig_ipv4);
+	if (err < 0)
+		goto err;
+
+	err = __connman_ipconfig_gateway_add(ipconfig_ipv4);
+	if (err < 0)
+		goto err;
+
+	__connman_service_indicate_state(service, CONNMAN_SERVICE_STATE_READY,
+						CONNMAN_IPCONFIG_TYPE_IPV4);
+
+	return;
+
+err:
+	connman_network_set_error(network,
+				CONNMAN_NETWORK_ERROR_CONFIGURE_FAIL);
+}
+
+static void dhcp_failure(struct connman_network *network)
+{
+	struct connman_service *service;
+	struct connman_ipconfig *ipconfig_ipv4;
+
+	service = __connman_service_lookup_from_network(network);
+	if (service == NULL)
+		return;
+
+	ipconfig_ipv4 = __connman_service_get_ip4config(service);
+	__connman_ipconfig_address_remove(ipconfig_ipv4);
+
+	__connman_service_indicate_state(service, CONNMAN_SERVICE_STATE_IDLE,
+						CONNMAN_IPCONFIG_TYPE_IPV4);
+}
+
+static void dhcp_callback(struct connman_network *network,
+			connman_bool_t success)
+{
+	DBG("success %d", success);
+
+	if (success == TRUE)
+		dhcp_success(network);
+	else
+		dhcp_failure(network);
 }
 
 static int set_connected_fixed(struct connman_network *network)
 {
 	struct connman_service *service;
-	struct connman_element *parent, *element;
+	struct connman_ipconfig *ipconfig_ipv4;
+	int err;
 
 	DBG("");
 
 	service = __connman_service_lookup_from_network(network);
 
-	parent = connman_network_get_element(network);
+	ipconfig_ipv4 = __connman_service_get_ip4config(service);
 
 	set_configuration(network);
 
-	if (parent->ipv4.address == NULL)
-		return -EINVAL;
-
-	if (parent->ipv4.netmask == NULL)
-		return -EINVAL;
-
-	element = connman_element_create(NULL);
-	if (element == NULL) {
-		connman_error("Can not create connman_element");
-		return -ENOMEM;
-	}
-
-	element->type = CONNMAN_ELEMENT_TYPE_IPV4;
-	element->index = parent->index;
-
-	if (connman_element_register(element, parent) < 0) {
-		connman_error("Can not register connman_element");
-		return -EINVAL;
-	}
-
 	network->connecting = FALSE;
 
 	connman_network_set_associating(network, FALSE);
 
+	err = __connman_ipconfig_address_add(ipconfig_ipv4);
+	if (err < 0)
+		goto err;
+
+	err = __connman_ipconfig_gateway_add(ipconfig_ipv4);
+	if (err < 0)
+		goto err;
+
+	__connman_service_indicate_state(service,
+					CONNMAN_SERVICE_STATE_READY,
+					CONNMAN_IPCONFIG_TYPE_IPV4);
 	return 0;
+
+err:
+	connman_network_set_error(network,
+			CONNMAN_NETWORK_ERROR_CONFIGURE_FAIL);
+
+	return err;
 }
 
 static void set_connected_manual(struct connman_network *network)
 {
 	struct connman_service *service;
 	struct connman_ipconfig *ipconfig;
-	const char *nameserver = NULL;
 	int err;
 
 	DBG("network %p", network);
@@ -659,49 +751,43 @@ static void set_connected_manual(struct connman_network *network)
 
 	set_configuration(network);
 
-	err = __connman_ipconfig_set_address(ipconfig);
-	if (err < 0) {
-		connman_network_set_error(network,
-			CONNMAN_NETWORK_ERROR_CONFIGURE_FAIL);
-		return;
-	}
+	err = __connman_ipconfig_address_add(ipconfig);
+	if (err < 0)
+		goto err;
 
-	connman_element_get_value(&network->element,
-			CONNMAN_PROPERTY_ID_IPV4_NAMESERVER, &nameserver);
-	if (nameserver != NULL)
-		__connman_service_append_nameserver(service, nameserver);
-
-	__connman_ipconfig_set_gateway(ipconfig, &network->element);
+	err = __connman_ipconfig_gateway_add(ipconfig);
+	if (err < 0)
+		goto err;
 
 	network->connecting = FALSE;
 
 	connman_network_set_associating(network, FALSE);
 
-	__connman_service_indicate_state(service, CONNMAN_SERVICE_STATE_READY);
+	__connman_service_indicate_state(service, CONNMAN_SERVICE_STATE_READY,
+					CONNMAN_IPCONFIG_TYPE_IPV4);
+
+	return;
+
+err:
+	connman_network_set_error(network,
+					CONNMAN_NETWORK_ERROR_CONFIGURE_FAIL);
+	return;
 }
 
 static int set_connected_dhcp(struct connman_network *network)
 {
-	struct connman_element *element;
-	int error;
+	int err;
 
 	DBG("network %p", network);
 
-	element = connman_element_create(NULL);
-	if (element == NULL)
-		return -ENOMEM;
-
-	element->type  = CONNMAN_ELEMENT_TYPE_DHCP;
-	element->index = network->element.index;
-
-	error = connman_element_register(element, &network->element);
-	if (error < 0) {
-		connman_element_unref(element);
-		return error;
-	}
-
 	set_configuration(network);
 
+	err = __connman_dhcp_start(network, dhcp_callback);
+	if (err < 0) {
+		connman_error("Can not request DHCP lease");
+		return err;
+	}
+
 	return 0;
 }
 
@@ -715,7 +801,7 @@ static int manual_ipv6_set(struct connman_network *network,
 	if (service == NULL)
 		return -EINVAL;
 
-	err = __connman_ipconfig_set_address(ipconfig_ipv6);
+	err = __connman_ipconfig_address_add(ipconfig_ipv6);
 	if (err < 0) {
 		connman_network_set_error(network,
 			CONNMAN_NETWORK_ERROR_CONFIGURE_FAIL);
@@ -730,6 +816,36 @@ static int manual_ipv6_set(struct connman_network *network,
 	return 0;
 }
 
+static void autoconf_ipv6_set(struct connman_network *network)
+{
+	struct connman_service *service;
+	struct connman_ipconfig *ipconfig;
+	const char *nameserver = NULL;
+
+	DBG("network %p", network);
+
+	service = __connman_service_lookup_from_network(network);
+
+	ipconfig = __connman_service_get_ip6config(service);
+
+	__connman_device_increase_connections(network->device);
+
+	__connman_device_set_network(network->device, network);
+
+	connman_device_set_disconnected(network->device, FALSE);
+
+	connman_element_get_value(&network->element,
+			CONNMAN_PROPERTY_ID_IPV6_NAMESERVER, &nameserver);
+	if (nameserver != NULL)
+		__connman_service_nameserver_append(service, nameserver);
+
+	network->connecting = FALSE;
+
+	__connman_service_indicate_state(service,
+					CONNMAN_SERVICE_STATE_READY,
+					CONNMAN_IPCONFIG_TYPE_IPV6);
+}
+
 static gboolean set_connected(gpointer user_data)
 {
 	struct connman_network *network = user_data;
@@ -757,7 +873,9 @@ static gboolean set_connected(gpointer user_data)
 		switch (ipv6_method) {
 		case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
 		case CONNMAN_IPCONFIG_METHOD_OFF:
+			break;
 		case CONNMAN_IPCONFIG_METHOD_AUTO:
+			autoconf_ipv6_set(network);
 			break;
 		case CONNMAN_IPCONFIG_METHOD_FIXED:
 		case CONNMAN_IPCONFIG_METHOD_MANUAL:
@@ -806,7 +924,22 @@ static gboolean set_connected(gpointer user_data)
 		service = __connman_service_lookup_from_network(network);
 
 		__connman_service_indicate_state(service,
-						CONNMAN_SERVICE_STATE_IDLE);
+					CONNMAN_SERVICE_STATE_DISCONNECT,
+					CONNMAN_IPCONFIG_TYPE_IPV4);
+
+		__connman_service_indicate_state(service,
+					CONNMAN_SERVICE_STATE_DISCONNECT,
+					CONNMAN_IPCONFIG_TYPE_IPV6);
+
+		__connman_connection_gateway_remove(service);
+
+		__connman_service_indicate_state(service,
+					CONNMAN_SERVICE_STATE_IDLE,
+					CONNMAN_IPCONFIG_TYPE_IPV4);
+
+		__connman_service_indicate_state(service,
+					CONNMAN_SERVICE_STATE_IDLE,
+					CONNMAN_IPCONFIG_TYPE_IPV6);
 	}
 
 	network->connecting = FALSE;
@@ -955,39 +1088,6 @@ int __connman_network_disconnect(struct connman_network *network)
 	return err;
 }
 
-static int dhcp_start(struct connman_network *network)
-{
-	struct connman_element *element;
-	int error;
-
-	element = connman_element_create(NULL);
-	if (element == NULL)
-		return -ENOMEM;
-
-	element->type  = CONNMAN_ELEMENT_TYPE_DHCP;
-	element->index = network->element.index;
-
-	error = connman_element_register(element, &network->element);
-	if (error < 0) {
-		connman_element_unref(element);
-		return error;
-	}
-
-	return 0;
-}
-
-static int dhcp_stop(struct connman_network *network)
-{
-	connman_element_unregister_children_type(&network->element,
-					CONNMAN_ELEMENT_TYPE_CONNECTION);
-	connman_element_unregister_children_type(&network->element,
-						CONNMAN_ELEMENT_TYPE_IPV4);
-	connman_element_unregister_children_type(&network->element,
-						CONNMAN_ELEMENT_TYPE_DHCP);
-
-	return 0;
-}
-
 static int manual_ipv4_set(struct connman_network *network,
 				struct connman_ipconfig *ipconfig)
 {
@@ -998,16 +1098,17 @@ static int manual_ipv4_set(struct connman_network *network,
 	if (service == NULL)
 		return -EINVAL;
 
-	err = __connman_ipconfig_set_address(ipconfig);
+	err = __connman_ipconfig_address_add(ipconfig);
 	if (err < 0) {
 		connman_network_set_error(network,
 			CONNMAN_NETWORK_ERROR_CONFIGURE_FAIL);
 		return err;
 	}
 
-	__connman_ipconfig_set_gateway(ipconfig, &network->element);
+	__connman_ipconfig_gateway_add(ipconfig);
 
-	__connman_service_indicate_state(service, CONNMAN_SERVICE_STATE_READY);
+	__connman_service_indicate_state(service, CONNMAN_SERVICE_STATE_READY,
+					CONNMAN_IPCONFIG_TYPE_IPV4);
 
 	return 0;
 }
@@ -1017,12 +1118,14 @@ int __connman_network_clear_ipconfig(struct connman_network *network,
 {
 	struct connman_service *service;
 	enum connman_ipconfig_method method;
+	enum connman_ipconfig_type type;
 
 	service = __connman_service_lookup_from_network(network);
 	if (service == NULL)
 		return -EINVAL;
 
 	method = __connman_ipconfig_get_method(ipconfig);
+	type = __connman_ipconfig_get_config_type(ipconfig);
 
 	switch (method) {
 	case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
@@ -1031,17 +1134,21 @@ int __connman_network_clear_ipconfig(struct connman_network *network,
 	case CONNMAN_IPCONFIG_METHOD_AUTO:
 		return -EINVAL;
 	case CONNMAN_IPCONFIG_METHOD_MANUAL:
-		connman_element_unregister_children_type(&network->element,
-					CONNMAN_ELEMENT_TYPE_CONNECTION);
-		__connman_ipconfig_clear_address(ipconfig);
+		__connman_ipconfig_address_remove(ipconfig);
 		break;
 	case CONNMAN_IPCONFIG_METHOD_DHCP:
-		dhcp_stop(network);
+		__connman_dhcp_stop(network);
 		break;
 	}
 
-	__connman_service_indicate_state(service,
-					CONNMAN_SERVICE_STATE_CONFIGURATION);
+	if (type == CONNMAN_IPCONFIG_TYPE_IPV6)
+		__connman_service_indicate_state(service,
+					CONNMAN_SERVICE_STATE_CONFIGURATION,
+					CONNMAN_IPCONFIG_TYPE_IPV6);
+	else if (type == CONNMAN_IPCONFIG_TYPE_IPV4)
+		__connman_service_indicate_state(service,
+					CONNMAN_SERVICE_STATE_CONFIGURATION,
+					CONNMAN_IPCONFIG_TYPE_IPV4);
 
 	return 0;
 }
@@ -1059,7 +1166,9 @@ int __connman_network_set_ipconfig(struct connman_network *network,
 		switch (method) {
 		case CONNMAN_IPCONFIG_METHOD_UNKNOWN:
 		case CONNMAN_IPCONFIG_METHOD_OFF:
+			break;
 		case CONNMAN_IPCONFIG_METHOD_AUTO:
+			autoconf_ipv6_set(network);
 			break;
 		case CONNMAN_IPCONFIG_METHOD_FIXED:
 		case CONNMAN_IPCONFIG_METHOD_MANUAL:
@@ -1087,43 +1196,96 @@ int __connman_network_set_ipconfig(struct connman_network *network,
 		case CONNMAN_IPCONFIG_METHOD_MANUAL:
 			return manual_ipv4_set(network, ipconfig_ipv4);
 		case CONNMAN_IPCONFIG_METHOD_DHCP:
-			return dhcp_start(network);
+			return __connman_dhcp_start(network, dhcp_callback);
 		}
 	}
 
 	return 0;
 }
 
-/**
- * connman_network_set_address:
- * @network: network structure
- * @address: binary address value
- * @size: binary address length
- *
- * Set unique address value for network
- */
-int connman_network_set_address(struct connman_network *network,
-				const void *address, unsigned int size)
+int connman_network_set_ipaddress(struct connman_network *network,
+					struct connman_ipaddress *ipaddress)
 {
-	const unsigned char *addr_octet = address;
-	char *str;
+	struct connman_service *service;
+	struct connman_ipconfig *ipconfig = NULL;
 
-	DBG("network %p size %d", network, size);
+	DBG("network %p", network);
 
-	if (size != 6)
+	service = __connman_service_lookup_from_network(network);
+	if (service == NULL)
 		return -EINVAL;
 
-	str = g_strdup_printf("%02X:%02X:%02X:%02X:%02X:%02X",
-				addr_octet[0], addr_octet[1], addr_octet[2],
-				addr_octet[3], addr_octet[4], addr_octet[5]);
-	if (str == NULL)
-		return -ENOMEM;
+	ipconfig = __connman_service_get_ipconfig(service, ipaddress->family);
+	if (ipconfig == NULL)
+		return -EINVAL;
 
-	g_free(network->address);
-	network->address = str;
+	__connman_ipconfig_set_local(ipconfig, ipaddress->local);
+	__connman_ipconfig_set_peer(ipconfig, ipaddress->peer);
+	__connman_ipconfig_set_broadcast(ipconfig, ipaddress->broadcast);
+	__connman_ipconfig_set_prefixlen(ipconfig, ipaddress->prefixlen);
+	__connman_ipconfig_set_gateway(ipconfig, ipaddress->gateway);
 
-	return connman_element_set_string(&network->element,
-						"Address", network->address);
+	return 0;
+}
+
+int connman_network_set_pac(struct connman_network *network,
+				const char *pac)
+{
+	struct connman_service *service;
+
+	DBG("network %p pac %s", network, pac);
+
+	service = __connman_service_lookup_from_network(network);
+	if (service == NULL)
+		return -EINVAL;
+
+	__connman_service_set_pac(service, pac);
+
+	return 0;
+}
+
+int connman_network_set_nameservers(struct connman_network *network,
+				const char *nameservers)
+{
+	struct connman_service *service;
+	char **nameservers_array = NULL;
+	int i;
+
+	DBG("network %p nameservers %s", network, nameservers);
+
+	service = __connman_service_lookup_from_network(network);
+	if (service == NULL)
+		return -EINVAL;
+
+	__connman_service_nameserver_clear(service);
+
+	if (nameservers != NULL)
+		nameservers_array = g_strsplit(nameservers, " ", 0);
+
+	for (i = 0; nameservers_array[i] != NULL; i++) {
+		__connman_service_nameserver_append(service,
+						nameservers_array[i]);
+	}
+
+	g_strfreev(nameservers_array);
+
+	return 0;
+}
+
+int connman_network_set_domain(struct connman_network *network,
+				const char *domain)
+{
+	struct connman_service *service;
+
+	DBG("network %p domain %s", network, domain);
+
+	service = __connman_service_lookup_from_network(network);
+	if (service == NULL)
+		return -EINVAL;
+
+	__connman_service_set_domainname(service, domain);
+
+	return 0;
 }
 
 /**
@@ -1198,10 +1360,7 @@ int connman_network_set_string(struct connman_network *network,
 	if (g_strcmp0(key, "Name") == 0)
 		return connman_network_set_name(network, value);
 
-	if (g_str_equal(key, "Address") == TRUE) {
-		g_free(network->address);
-		network->address = g_strdup(value);
-	} else if (g_str_equal(key, "Node") == TRUE) {
+	if (g_str_equal(key, "Node") == TRUE) {
 		g_free(network->node);
 		network->node = g_strdup(value);
 	} else if (g_str_equal(key, "WiFi.Mode") == TRUE) {
@@ -1234,6 +1393,9 @@ int connman_network_set_string(struct connman_network *network,
 	} else if (g_str_equal(key, "WiFi.Phase2") == TRUE) {
 		g_free(network->wifi.phase2_auth);
 		network->wifi.phase2_auth = g_strdup(value);
+	} else if (g_str_equal(key, "WiFi.PinWPS") == TRUE) {
+		g_free(network->wifi.pin_wps);
+		network->wifi.pin_wps = g_strdup(value);
 	}
 
 	err = connman_element_set_string(&network->element, key, value);
@@ -1261,9 +1423,7 @@ const char *connman_network_get_string(struct connman_network *network,
 {
 	DBG("network %p key %s", network, key);
 
-	if (g_str_equal(key, "Address") == TRUE)
-		return network->address;
-	else if (g_str_equal(key, "Name") == TRUE)
+	if (g_str_equal(key, "Name") == TRUE)
 		return network->name;
 	else if (g_str_equal(key, "Node") == TRUE)
 		return network->node;
@@ -1287,6 +1447,8 @@ const char *connman_network_get_string(struct connman_network *network,
 		return network->wifi.private_key_passphrase;
 	else if (g_str_equal(key, "WiFi.Phase2") == TRUE)
 		return network->wifi.phase2_auth;
+	else if (g_str_equal(key, "WiFi.PinWPS") == TRUE)
+		return network->wifi.pin_wps;
 
 	return connman_element_get_string(&network->element, key);
 }
@@ -1306,6 +1468,10 @@ int connman_network_set_bool(struct connman_network *network,
 
 	if (g_strcmp0(key, "Roaming") == 0)
 		return connman_network_set_roaming(network, value);
+	else if (g_strcmp0(key, "WiFi.WPS") == 0)
+		network->wifi.wps = value;
+	else if (g_strcmp0(key, "WiFi.UseWPS") == 0)
+		network->wifi.use_wps = value;
 
 	return connman_element_set_bool(&network->element, key, value);
 }
@@ -1324,6 +1490,10 @@ connman_bool_t connman_network_get_bool(struct connman_network *network,
 
 	if (g_str_equal(key, "Roaming") == TRUE)
 		return network->roaming;
+	else if (g_str_equal(key, "WiFi.WPS") == TRUE)
+		return network->wifi.wps;
+	else if (g_str_equal(key, "WiFi.UseWPS") == TRUE)
+		return network->wifi.use_wps;
 
 	return connman_element_get_bool(&network->element, key);
 }
@@ -1420,9 +1590,6 @@ int connman_network_set_blob(struct connman_network *network,
 {
 	DBG("network %p key %s size %d", network, key, size);
 
-	if (g_strcmp0(key, "Address") == 0)
-		return connman_network_set_address(network, data, size);
-
 	if (g_str_equal(key, "WiFi.SSID") == TRUE) {
 		g_free(network->wifi.ssid);
 		network->wifi.ssid = g_try_malloc(size);
