@@ -25,6 +25,7 @@
 
 #include <stdio.h>
 #include <string.h>
+#include <stdlib.h>
 #include <gdbus.h>
 
 #include "connman.h"
@@ -36,6 +37,7 @@ static GHashTable *provider_hash = NULL;
 static GSList *driver_list = NULL;
 
 struct connman_route {
+	int family;
 	char *host;
 	char *netmask;
 	char *gateway;
@@ -48,9 +50,9 @@ struct connman_provider {
 	char *name;
 	char *type;
 	char *host;
-	char *dns;
 	char *domain;
-	GSList *route_list;
+	int family;
+	GHashTable *routes;
 	struct connman_provider_driver *driver;
 	void *driver_data;
 };
@@ -80,35 +82,6 @@ static struct connman_provider *connman_provider_lookup(const char *identifier)
 	return provider;
 }
 
-static int connman_provider_setup_vpn_ipv4(struct connman_provider *provider,
-						struct connman_element *element)
-{
-	if (element == NULL || provider == NULL)
-		return -EINVAL;
-
-	DBG("set vpn type %d", element->type);
-
-	g_free(element->ipv4.address);
-	element->ipv4.address = g_strdup(provider->element.ipv4.address);
-
-	g_free(element->ipv4.peer);
-	element->ipv4.peer = g_strdup(provider->element.ipv4.peer);
-
-	g_free(element->ipv4.netmask);
-	element->ipv4.netmask = g_strdup(provider->element.ipv4.netmask);
-
-	g_free(element->ipv4.gateway);
-	element->ipv4.gateway = g_strdup(provider->element.ipv4.gateway);
-
-	g_free(element->ipv4.broadcast);
-	element->ipv4.broadcast = g_strdup(provider->element.ipv4.broadcast);
-
-	g_free(element->ipv4.pac);
-	element->ipv4.pac = g_strdup(provider->element.ipv4.pac);
-
-	return connman_element_register(element, &provider->element);
-}
-
 struct connman_provider *connman_provider_ref(struct connman_provider *provider)
 {
 	DBG("provider %p", provider);
@@ -164,6 +137,18 @@ static int provider_probe(struct connman_provider *provider)
 	return 0;
 }
 
+static int provider_indicate_state(struct connman_provider *provider,
+					enum connman_service_state state)
+{
+	DBG("state %d", state);
+
+	__connman_service_indicate_state(provider->vpn_service, state,
+					CONNMAN_IPCONFIG_TYPE_IPV4);
+
+	return __connman_service_indicate_state(provider->vpn_service, state,
+					CONNMAN_IPCONFIG_TYPE_IPV6);
+}
+
 int __connman_provider_disconnect(struct connman_provider *provider)
 {
 	int err;
@@ -176,8 +161,9 @@ int __connman_provider_disconnect(struct connman_provider *provider)
 		return -EOPNOTSUPP;
 
 	if (provider->vpn_service != NULL)
-		__connman_service_indicate_state(provider->vpn_service,
+		provider_indicate_state(provider,
 					CONNMAN_SERVICE_STATE_DISCONNECT);
+
 	if (err < 0) {
 		if (err != -EINPROGRESS)
 			return err;
@@ -194,20 +180,6 @@ int __connman_provider_connect(struct connman_provider *provider)
 
 	DBG("provider %p", provider);
 
-	g_free(provider->element.ipv4.address);
-	g_free(provider->element.ipv4.peer);
-	g_free(provider->element.ipv4.netmask);
-	g_free(provider->element.ipv4.gateway);
-	g_free(provider->element.ipv4.broadcast);
-	g_free(provider->element.ipv4.pac);
-
-	provider->element.ipv4.address = NULL;
-	provider->element.ipv4.peer = NULL;
-	provider->element.ipv4.netmask = NULL;
-	provider->element.ipv4.gateway = NULL;
-	provider->element.ipv4.broadcast = NULL;
-	provider->element.ipv4.pac = NULL;
-
 	if (provider->driver != NULL && provider->driver->connect != NULL)
 		err = provider->driver->connect(provider);
 	else
@@ -217,8 +189,9 @@ int __connman_provider_connect(struct connman_provider *provider)
 		if (err != -EINPROGRESS)
 			return err;
 
-		__connman_service_indicate_state(provider->vpn_service,
+		provider_indicate_state(provider,
 					CONNMAN_SERVICE_STATE_ASSOCIATION);
+
 		return -EINPROGRESS;
 	}
 
@@ -254,85 +227,62 @@ int __connman_provider_remove(const char *path)
 	return -ENXIO;
 }
 
+static void provider_append_routes(gpointer key, gpointer value,
+					gpointer user_data)
+{
+	struct connman_route *route = value;
+	struct connman_provider *provider = user_data;
+	int index = provider->element.index;
+
+	if (route->family == AF_INET6) {
+		unsigned char prefix_len = atoi(route->netmask);
+
+		connman_inet_add_ipv6_network_route(index, route->host,
+							route->gateway,
+							prefix_len);
+	} else {
+		connman_inet_add_network_route(index, route->host,
+						route->gateway,
+						route->netmask);
+	}
+}
+
 static int set_connected(struct connman_provider *provider,
 					connman_bool_t connected)
 {
 	struct connman_service *service = provider->vpn_service;
+	struct connman_ipconfig *ipconfig;
 
 	if (service == NULL)
 		return -ENODEV;
 
+	ipconfig = __connman_service_get_ipconfig(service, provider->family);
+
 	if (connected == TRUE) {
-		enum connman_element_type type = CONNMAN_ELEMENT_TYPE_UNKNOWN;
-		struct connman_element *element;
-		char *nameservers = NULL, *name = NULL;
-		const char *value;
-		char *second_ns;
-		GSList *list;
-		int err;
-
-		__connman_service_indicate_state(provider->vpn_service,
-					CONNMAN_SERVICE_STATE_CONFIGURATION);
-
-		type = CONNMAN_ELEMENT_TYPE_IPV4;
-
-		element = connman_element_create(NULL);
-		if (element == NULL)
-			return -ENOMEM;
-
-		element->type  = type;
-		element->index = provider->element.index;
-
-		err = connman_provider_setup_vpn_ipv4(provider, element);
-		if (err < 0) {
-			connman_element_unref(element);
-
-			__connman_service_indicate_state(service,
+		if (ipconfig == NULL) {
+			provider_indicate_state(provider,
 						CONNMAN_SERVICE_STATE_FAILURE);
-
-			return err;
+			return -EIO;
 		}
 
-		__connman_service_indicate_state(service,
-						CONNMAN_SERVICE_STATE_READY);
+		__connman_ipconfig_address_add(ipconfig);
+		__connman_ipconfig_gateway_add(ipconfig);
 
-		__connman_service_set_domainname(service, provider->domain);
+		provider_indicate_state(provider,
+					CONNMAN_SERVICE_STATE_READY);
 
-		name = connman_inet_ifname(provider->element.index);
+		g_hash_table_foreach(provider->routes, provider_append_routes,
+					provider);
 
-		nameservers = g_strdup(provider->dns);
-		value = nameservers;
-		second_ns = strchr(value, ' ');
-		if (second_ns)
-			*(second_ns++) = 0;
-		__connman_service_append_nameserver(service, value);
-		value = second_ns;
-
-		while (value) {
-			char *next = strchr(value, ' ');
-			if (next)
-				*(next++) = 0;
-
-			connman_resolver_append(name, provider->domain, value);
-			value = next;
-		}
-
-		g_free(nameservers);
-		g_free(name);
-
-		for (list = provider->route_list; list; list = list->next) {
-			struct connman_route *route = list->data;
-
-			connman_inet_add_network_route(provider->element.index,
-							route->host,
-							route->gateway,
-							route->netmask);
-			/* XXX Need to handle IPv6 too */
-		}
 	} else {
-		connman_element_unregister_children(&provider->element);
-		__connman_service_indicate_state(service,
-						CONNMAN_SERVICE_STATE_IDLE);
+		if (ipconfig != NULL) {
+			provider_indicate_state(provider,
+					CONNMAN_SERVICE_STATE_DISCONNECT);
+			__connman_ipconfig_gateway_remove(ipconfig);
+		}
+
+		provider_indicate_state(provider,
+					CONNMAN_SERVICE_STATE_IDLE);
 	}
 
 	return 0;
@@ -350,21 +300,45 @@ int connman_provider_set_state(struct connman_provider *provider,
 	case CONNMAN_PROVIDER_STATE_IDLE:
 		return set_connected(provider, FALSE);
 	case CONNMAN_PROVIDER_STATE_CONNECT:
-		return __connman_service_indicate_state(provider->vpn_service,
+		return provider_indicate_state(provider,
 					CONNMAN_SERVICE_STATE_ASSOCIATION);
 	case CONNMAN_PROVIDER_STATE_READY:
 		return set_connected(provider, TRUE);
 	case CONNMAN_PROVIDER_STATE_DISCONNECT:
-		return __connman_service_indicate_state(provider->vpn_service,
+		return provider_indicate_state(provider,
 					CONNMAN_SERVICE_STATE_DISCONNECT);
 	case CONNMAN_PROVIDER_STATE_FAILURE:
-		return __connman_service_indicate_state(provider->vpn_service,
+		return provider_indicate_state(provider,
 					CONNMAN_SERVICE_STATE_FAILURE);
 	}
 
 	return -EINVAL;
 }
 
+int connman_provider_indicate_error(struct connman_provider *provider,
+					enum connman_provider_error error)
+{
+	enum connman_service_error service_error;
+
+	switch (error) {
+	case CONNMAN_PROVIDER_ERROR_LOGIN_FAILED:
+		service_error = CONNMAN_SERVICE_ERROR_LOGIN_FAILED;
+		break;
+	case CONNMAN_PROVIDER_ERROR_AUTH_FAILED:
+		service_error = CONNMAN_SERVICE_ERROR_AUTH_FAILED;
+		break;
+	case CONNMAN_PROVIDER_ERROR_CONNECT_FAILED:
+		service_error = CONNMAN_SERVICE_ERROR_CONNECT_FAILED;
+		break;
+	default:
+		service_error = CONNMAN_SERVICE_ERROR_UNKNOWN;
+		break;
+	}
+
+	return __connman_service_indicate_error(provider->vpn_service,
+							service_error);
+}
+
 static void unregister_provider(gpointer data)
 {
 	struct connman_provider *provider = data;
@@ -382,7 +356,6 @@ static void unregister_provider(gpointer data)
 static void provider_destruct(struct connman_element *element)
 {
 	struct connman_provider *provider = element->private;
-	GSList *list;
 
 	DBG("provider %p", provider);
 
@@ -390,16 +363,17 @@ static void provider_destruct(struct connman_element *element)
 	g_free(provider->type);
 	g_free(provider->domain);
 	g_free(provider->identifier);
-	g_free(provider->dns);
+	g_hash_table_destroy(provider->routes);
+}
 
-	for (list = provider->route_list; list; list = list->next) {
-		struct connman_route *route = list->data;
+static void destroy_route(gpointer user_data)
+{
+	struct connman_route *route = user_data;
 
-		g_free(route->host);
-		g_free(route->netmask);
-		g_free(route->gateway);
-	}
-	g_slist_free(provider->route_list);
+	g_free(route->host);
+	g_free(route->netmask);
+	g_free(route->gateway);
+	g_free(route);
 }
 
 static void provider_initialize(struct connman_provider *provider)
@@ -411,18 +385,12 @@ static void provider_initialize(struct connman_provider *provider)
 	provider->element.private = provider;
 	provider->element.destruct = provider_destruct;
 
-	provider->element.ipv4.address = NULL;
-	provider->element.ipv4.netmask = NULL;
-	provider->element.ipv4.gateway = NULL;
-	provider->element.ipv4.broadcast = NULL;
-	provider->element.ipv4.pac = NULL;
-
 	provider->name = NULL;
 	provider->type = NULL;
-	provider->dns = NULL;
 	provider->domain = NULL;
 	provider->identifier = NULL;
-	provider->route_list = NULL;
+	provider->routes = g_hash_table_new_full(g_direct_hash, g_direct_equal,
+					NULL, destroy_route);
 }
 
 static struct connman_provider *connman_provider_new(void)
@@ -636,29 +604,6 @@ int connman_provider_set_string(struct connman_provider *provider,
 	} else if (g_str_equal(key, "Name") == TRUE) {
 		g_free(provider->name);
 		provider->name = g_strdup(value);
-	} else if (g_str_equal(key, "Gateway") == TRUE) {
-		g_free(provider->element.ipv4.gateway);
-		provider->element.ipv4.gateway = g_strdup(value);
-	} else if (g_str_equal(key, "Address") == TRUE) {
-		g_free(provider->element.ipv4.address);
-		provider->element.ipv4.address = g_strdup(value);
-	} else if (g_str_equal(key, "Peer") == TRUE) {
-		g_free(provider->element.ipv4.peer);
-		provider->element.ipv4.peer = g_strdup(value);
-	} else if (g_str_equal(key, "Netmask") == TRUE) {
-		g_free(provider->element.ipv4.netmask);
-		provider->element.ipv4.netmask = g_strdup(value);
-	} else if (g_str_equal(key, "PAC") == TRUE) {
-		g_free(provider->element.ipv4.pac);
-		provider->element.ipv4.pac = g_strdup(value);
-		__connman_service_set_proxy_autoconfig(provider->vpn_service,
-									value);
-	} else if (g_str_equal(key, "DNS") == TRUE) {
-		g_free(provider->dns);
-		provider->dns = g_strdup(value);
-	} else if (g_str_equal(key, "Domain") == TRUE) {
-		g_free(provider->domain);
-		provider->domain = g_strdup(value);
 	}
 
 	return connman_element_set_string(&provider->element, key, value);
@@ -737,23 +682,178 @@ int connman_provider_get_index(struct connman_provider *provider)
 	return provider->element.index;
 }
 
+int connman_provider_set_ipaddress(struct connman_provider *provider,
+					struct connman_ipaddress *ipaddress)
+{
+	struct connman_ipconfig *ipconfig = NULL;
+
+	ipconfig = __connman_service_get_ipconfig(provider->vpn_service,
+							ipaddress->family);
+	if (ipconfig == NULL)
+		return -EINVAL;
+
+	provider->family = ipaddress->family;
+
+	__connman_ipconfig_set_local(ipconfig, ipaddress->local);
+	__connman_ipconfig_set_peer(ipconfig, ipaddress->peer);
+	__connman_ipconfig_set_broadcast(ipconfig, ipaddress->broadcast);
+	__connman_ipconfig_set_gateway(ipconfig, ipaddress->gateway);
+	__connman_ipconfig_set_prefixlen(ipconfig, ipaddress->prefixlen);
+
+	return 0;
+}
+
+int connman_provider_set_pac(struct connman_provider *provider, const char *pac)
+{
+	DBG("provider %p pac %s", provider, pac);
+
+	__connman_service_set_pac(provider->vpn_service, pac);
+
+	return 0;
+}
+
+
+int connman_provider_set_domain(struct connman_provider *provider,
+					const char *domain)
+{
+	DBG("provider %p domain %s", provider, domain);
+
+	g_free(provider->domain);
+	provider->domain = g_strdup(domain);
+
+	__connman_service_set_domainname(provider->vpn_service, domain);
+
+	return 0;
+}
+
+int connman_provider_set_nameservers(struct connman_provider *provider,
+					const char *nameservers)
+{
+	int i;
+	char **nameservers_array = NULL;
+
+	DBG("provider %p nameservers %s", provider, nameservers);
+
+	__connman_service_nameserver_clear(provider->vpn_service);
+
+	if (nameservers == NULL)
+		return 0;
+
+	nameservers_array = g_strsplit(nameservers, " ", 0);
+
+	for (i = 0; nameservers_array[i] != NULL; i++) {
+		__connman_service_nameserver_append(provider->vpn_service,
+							nameservers_array[i]);
+	}
+
+	g_strfreev(nameservers_array);
+
+	return 0;
+}
+
+enum provider_route_type {
+	PROVIDER_ROUTE_TYPE_NONE = 0,
+	PROVIDER_ROUTE_TYPE_MASK = 1,
+	PROVIDER_ROUTE_TYPE_ADDR = 2,
+	PROVIDER_ROUTE_TYPE_GW   = 3,
+};
+
+static int route_env_parse(struct connman_provider *provider, const char *key,
+				int *family, unsigned long *idx,
+				enum provider_route_type *type)
+{
+	char *end;
+	const char *start;
+
+	DBG("name %s", provider->name);
+
+	if (!strcmp(provider->type, "openvpn")) {
+		if (g_str_has_prefix(key, "route_network_") == TRUE) {
+			start = key + strlen("route_network_");
+			*type = PROVIDER_ROUTE_TYPE_ADDR;
+		} else if (g_str_has_prefix(key, "route_netmask_") == TRUE) {
+			start = key + strlen("route_netmask_");
+			*type = PROVIDER_ROUTE_TYPE_MASK;
+		} else if (g_str_has_prefix(key, "route_gateway_") == TRUE) {
+			start = key + strlen("route_gateway_");
+			*type = PROVIDER_ROUTE_TYPE_GW;
+		} else
+			return -EINVAL;
+
+		*family = AF_INET;
+		*idx = g_ascii_strtoull(start, &end, 10);
+
+	} else if (!strcmp(provider->type, "openconnect")) {
+		if (g_str_has_prefix(key, "CISCO_SPLIT_INC_") == TRUE) {
+			*family = AF_INET;
+			start = key + strlen("CISCO_SPLIT_INC_");
+		} else if (g_str_has_prefix(key, "CISCO_IPV6_SPLIT_INC_") == TRUE) {
+			*family = AF_INET6;
+			start = key + strlen("CISCO_IPV6_SPLIT_INC_");
+		} else
+			return -EINVAL;
+
+		*idx = g_ascii_strtoull(start, &end, 10);
+
+		if (strncmp(end, "_ADDR", 5) == 0)
+			*type = PROVIDER_ROUTE_TYPE_ADDR;
+		else if (strncmp(end, "_MASK", 5) == 0)
+			*type = PROVIDER_ROUTE_TYPE_MASK;
+		else if (strncmp(end, "_MASKLEN", 8) == 0 &&
+				*family == AF_INET6) {
+			*type = PROVIDER_ROUTE_TYPE_MASK;
+		} else
+			return -EINVAL;
+	}
+
+	return 0;
+}
+
 int connman_provider_append_route(struct connman_provider *provider,
-					const char *host, const char *netmask,
-					const char *gateway)
+					const char *key, const char *value)
 {
 	struct connman_route *route;
+	int ret, family = 0;
+	unsigned long idx = 0;
+	enum provider_route_type type = PROVIDER_ROUTE_TYPE_NONE;
 
-	route = g_try_new0(struct connman_route, 1);
-	if (route == NULL)
-		return -ENOMEM;
+	DBG("key %s value %s", key, value);
 
-	route->host = g_strdup(host);
-	route->netmask = g_strdup(netmask);
-	route->gateway = g_strdup(gateway);
+	ret = route_env_parse(provider, key, &family, &idx, &type);
+	if (ret < 0)
+		return ret;
 
-	provider->route_list = g_slist_append(provider->route_list, route);
+	DBG("idx %lu family %d type %d", idx, family, type);
 
-	return TRUE;
+	route = g_hash_table_lookup(provider->routes, GINT_TO_POINTER(idx));
+	if (route == NULL) {
+		route = g_try_new0(struct connman_route, 1);
+		if (route == NULL) {
+			connman_error("out of memory");
+			return -ENOMEM;
+		}
+
+		route->family = family;
+
+		g_hash_table_replace(provider->routes, GINT_TO_POINTER(idx),
+						route);
+	}
+
+	switch (type) {
+	case PROVIDER_ROUTE_TYPE_NONE:
+		break;
+	case PROVIDER_ROUTE_TYPE_MASK:
+		route->netmask = g_strdup(value);
+		break;
+	case PROVIDER_ROUTE_TYPE_ADDR:
+		route->host = g_strdup(value);
+		break;
+	case PROVIDER_ROUTE_TYPE_GW:
+		route->gateway = g_strdup(value);
+		break;
+	}
+
+	return 0;
 }
 
 const char *connman_provider_get_driver_name(struct connman_provider *provider)
